Attribute VB_Name = "SignalFilter"
Option Explicit


' Windowed Sinc Filter und Convolution
' nach http://www.dspguide.com/
'
' factor:   Wert von 0.0 bis 0.5, der den Cutoff angibt.
'           Bsp.: Audiosignal gesampelt mit 22050 Hz,
'                 Cutoff Frequenz soll bei 220.5 Hz liegen,
'                 dann wre factor = 220.5/22050 = 0.01
'
' Cutoff:   Schwelle, ab der das Filter beginnt zu wirken
'
' Taps:     Qualitt des Filters, je mehr Taps, desto steiler
'           das Rolloff Band nach dem Cutoff, aber je mehr Taps,
'           desto grer das Delay, und desto mehr Rechenarbeit


Public Type FilterKernel
    kernel()        As Single
    olap()          As Single
    taps            As Long
End Type

Public Enum FilterType
    FilterHighpass = 0
    FilterLowpass
End Enum

Private Const PI    As Single = 3.14159265358979
Private Const PI2   As Single = PI * 2

' Windows Sinc FIR Filter Kern berechnen
' nach http://www.dspguide.com/
Public Function CreateFilter(ByVal ftp As FilterType, ByVal taps As Long, _
    ByVal factor As Single) As FilterKernel
    
    Dim omega       As Single
    Dim n           As Single
    Dim sum         As Single
    Dim m           As Long
    Dim i           As Long

    omega = PI2 * factor
    m = taps / 2

    With CreateFilter
        ReDim .kernel(taps - 1) As Single
        ReDim .olap(taps - 1) As Single
        .taps = taps
    
        ' Sinc Funktion
        For i = 0 To taps - 1
            If i - m = 0 Then
                .kernel(i) = omega
            Else
                n = i - m
                .kernel(i) = Sin(omega * n) / n
            End If
            
            ' Glockenfenster um Ripple zu minimieren
            .kernel(i) = .kernel(i) * HammingWindow(i, taps)
            sum = sum + .kernel(i)
        Next
        
        If sum = 0 Then sum = 1
        
        For i = 0 To taps - 1
            ' Kern normalisieren
            .kernel(i) = .kernel(i) / sum
        Next
        
        If ftp = FilterHighpass Then
            ' Spektrale Inversion, um Lowpass in Highpass umzuwandeln
            For i = 0 To taps - 1
                .kernel(i) = -.kernel(i)
            Next
            .kernel(m) = .kernel(m) + 1
        End If
        
        '' durch Convolution des Kerns mit sich selbst
        '' kann das Stopband nochmal verstrkt werden,
        '' allerdings ergeben sich daraus auch mehr Taps
        ' Dim k2() As Single
        ' k2 = .kernel
        ' .kernel = Convolve(k2, .kernel)
        ' ReDim .olap(UBound(.kernel)) As Single
    End With
End Function

' Leert Overlap von Filter
Public Sub ResetFilter(kernel As FilterKernel)
    Dim i           As Long
    
    For i = 0 To kernel.taps - 1
        kernel.olap(i) = 0
    Next
End Sub

' Signalfaltung mit Filterkernel, Overlap-Add Methode
Public Sub FilterProcess(sngValues() As Single, kernel As FilterKernel)
    Dim sngOut()    As Single
    Dim i           As Long
    Dim n           As Long
    
    n = UBound(sngValues) + 1
    
    If n >= kernel.taps - 1 Then
        sngOut = Convolve(sngValues, kernel.kernel)

        For i = 0 To kernel.taps - 1
            sngOut(i) = sngOut(i) + kernel.olap(i)
            kernel.olap(i) = sngOut(n + i)
        Next

        For i = 0 To n - 1
            sngValues(i) = sngOut(i)
        Next
    End If
End Sub

' Faltung nach Input Side Methode
Private Function Convolve(a() As Single, B() As Single) As Single()
    Dim c() As Single
    Dim i   As Long
    Dim j   As Long
    
    ReDim c(UBound(a) + UBound(B) + 1) As Single
    
    For i = 0 To UBound(a)
        For j = 0 To UBound(B)
            c(i + j) = c(i + j) + a(i) * B(j)
        Next
    Next
    
    Convolve = c
End Function

' http://www.dspguide.com/
Private Function HammingWindow(ByVal i As Single, ByVal n As Single) As Single
    HammingWindow = 0.54 - 0.46 * Cos(PI2 * i / n)
End Function
